/*
 * The I3 router
 * This object serves 'user' objects for each connection, keeps track of them,
 * and allows sending data between them.
 *
 * It also supports the intermud 3 irn protocol.
 *
 * This is a long and windy piece of code, it should be split up.. not for today.
 *
 * Aidil @ Way of the Force.
 */

# include <config.h>
# include <limits.h>
# include <type.h>
# include <intermud3.h>
# include <i3spec.h>
# include <pathnames.h>
# include <loglevels.h>

inherit netlib  "/system/lib/net/net";
inherit mudlist "/system/lib/router/mudlist";
inherit irnsrv  "/system/lib/router/irnv1srv";

# include <router.h>                    /* also contains prototypes, needs to be after all inherit statements */

static mapping peers;                   /* mapping between peer name and their 'user' objects */
static object driver;                   /* driver object */
static int lasterror;                   /* last error we got */
static int lastclean;

string myname;                          /* name of this router */
string mypass;
string netname;                         /* name of our network, used to communicate with local intermud daemon */
mapping chandata;
mapping routerlist;                     /* router list */
int chanlistid;
int irnchanlistid;
int DBL;

void init_routerlist();

static void save_me() {
  save_object(SAVEFILE);
}

static int restore_me() {
  return restore_object(SAVEFILE);
}

void enable_debug(int flag) {
  DBL = DBL | flag;
  mudlist::set_mudlist_dbl(DBL);
  irnsrv::set_irn_dbl(DBL);
}

void disable_debug(int flag) {
  DBL = DBL & (0xFF ^ flag);
  mudlist::set_mudlist_dbl(DBL);
  irnsrv::set_irn_dbl(DBL);
}

int query_debug() {
  return DBL;
}

static void create() {
    netlib::create();
    driver = find_object(DRIVER);
    seteuid(getuid());
    if(!restore_me()) {
      myname = NAME;
      netname = "yatmim";
      init_routerlist();
    }

    mypass = MYPASS;

    mudlist::create();
    mudlist::set_myname(myname);
    mudlist::set_mudlist_dbl(DBL);
    save_me();

    set_handler(HANDLER);
    open_port("mudmode", PORT);   
    peers = ([]);
    irn = ([]);
    NETWORK->add_listener();
    call_out("start_client",5);
    call_out("periodic",60);
}

static mapping i3_nodes() {
  return peers;
}

static string *i3_node_names() {
  return m_indices(peers);
}

static object *i3_node_objects() {
  return m_values(peers);
}

static string query_mypass() { return mypass; }
static string query_myname() { return myname; }

static void start_client() {
  object ob;
  string * routers;
  int i;

  routers = m_indices(routerlist);

  for(i=0;i<sizeof(routers);i++) {
    if(routers[i] != myname && !irn[routers[i]]) {
      start_irn_con(routers[i]);
    }
  }
}

string query_nodename() { return myname; }
int query_localsrc() { return 1; }
int query_irn() { return 1; }

object connection(string ip, int port)
{
    return ::connection(allocate(driver->query_tls_size()), ip, port, "tcp");
}

void done() {
    close_user();
}

/*
 * NAME:	open()
 * DESCRIPTION:	open the connection
 */
static int open(int port) {
    ::open(allocate(driver->query_tls_size()), port);
    return 0;
}

/*
 * NAME:	close()
 * DESCRIPTION:	close the connection
 */
static void close(int force) {
    ::close(allocate(driver->query_tls_size()), force);
}

static void close_peer(object p) {
  p->force_disconnect();
}

static void destroying() {

  NETWORK->remove_listener();
  broadcast_packet( ({ "router-shutdown", 5, myname, 0, 0, 0, 60 }),0 );
/*
  filter_array(m_values(peers)+m_values(irn),"close_peer",this_object());
*/
  filter_array(HANDLER->query_clone_list(),"close_peer",this_object());
  NETWORK->remove_listener();
  IMUDWARN("Destructed router object\n");
  mudlist::destroying();
  save_me();
}

static void register_node(object o,string n) {
  if(!peers) peers = ([]);
  o->set_nodename(n);
  o->set_irn(0);
  o->set_localsrc(1);
  peers[n]=o;
  muddata[n]["lastip"] = query_ip_number(o);
  muddata[n]["router"] = myname;
  muddata[n]["delay"] = -1;
  IMUDLOG(n+" connected from "+query_ip_number(o));
}

static int authenticate_node(object con, mixed *data) {
  string src,dst;
  int version;
  mapping md;
  
  IMUDDEBUG(DEBUG_AUTH,"Authentication request from "+query_ip_number(con));
  IMUDDEBUG(DEBUG_DATA,sprintf("%y\n",data));
  if(sizeof(data) < 7) return ERROR_BAD_PACKET;

  src = (data[2] ? data[2]:nil);
  dst = (data[4] ? data[4]:nil);

  IMUDDEBUG(DEBUG_AUTH,"Claims to be "+(src ? src:"<NO NAME>"));

  if(!src) {
    return ERROR_BAD_SOURCE;
  }

  if(src == "Mudfarm Lima" && dst == "*wir") dst = "*war";

  if(dst != myname) {
    IMUDDEBUG(DEBUG_AUTH,"bad destination for auth packet: "+dst);
    return ERROR_INVALID_DESTINATION;
  }

  sscanf(data[0],"startup-req-%d",version);

  md = get_muddata(src);

  if(!md) {
    md = ([
      "router":myname,
      "password":random(INT_MAX),
      "connect_time":time(),
      "disconnect_time":0,
      "version":version,
      "lastip":query_ip_number(con),
      "public":data[7..],
      "delay":-1
    ]);
    IMUDDEBUG(DEBUG_AUTH,"registering new node "+src);
  } else {
    int pw;
    string ip;

    pw = md["password"];
#ifdef PUBLIC_ROUTER
    ip = query_ip_number(con);
    if(ip != md["lastip"] && data[6] != pw) {
#else
    if(data[6] != pw) {
#endif
      IMUDDEBUG(DEBUG_AUTH,"Bad password");
      IMUDERROR("Bad password from "+src+" ("+query_ip_number(previous_object())+")");
      return ERROR_PERMISSION_DENIED;
    } else {
      IMUDDEBUG(DEBUG_AUTH,"Authenticated node "+src);
      md = ([
      "router":myname,
      "connect_time":time(),
      "disconnect_time":0,
      "version":version,
      "lastip":query_ip_number(con),
      "public":data[7..],
      "password":pw,
      "delay":-1
      ]);

    }
  }
  store_muddata(src,md);
  irn_broadcast_mudlist_change(src);
  register_node(con,src);
  save_me();
}

static mapping query_router_list() {
  return routerlist ? routerlist[..]:([]);
}

static mixed * filtered_routerlist() {
  int i;
  mixed * ret, rnames;

  ret = ({ });

  rnames = m_indices(routerlist);

  for(i=0;i<sizeof(rnames);i++) {
    if(rnames[i] == myname) 
    ret += ({
      ({ 
        rnames[i],  
        routerlist[rnames[i]][0] + " " + routerlist[rnames[i]][1]
      })
    });
  }
  return ret;
}

object find_node(string n, varargs int route) {
  /*
   * This object is a node as well for directed requests
   */
  if(myname && n == myname) {
    return this_object();
  } else if(peers[n]) {
    return peers[n];
  } else if(route && muddata[n] && muddata[n]["delay"] == -1) {
    string rname;
    object con;
    rname = muddata[n]["router"];
    if(rname) con = find_irn(rname);
    return con;
  }
}

static int allow_packet(object con, mixed * packet) {
  /*
   * set lasterror to something. The code below will have to clear it
   */
  lasterror = ERROR_BAD_PACKET;

  if(!stringp(packet[0])) {
    lasterror = ERROR_BAD_PACKET;
  /*
   * refuse packets from bad sites outright.
   */
  } else if(NETWORK->query_bad_site(query_ip_number(con)) > 3) {
    IMUDERROR("packet from banned site "+query_ip_number(con));
    lasterror = ERROR_BANNED;
  /*
   * Authenticate startup-req-* packets
   */
  } else if(!con->query_nodename() && sscanf(packet[0],"startup-req-%*d") == 1) {
    lasterror = authenticate_node(con,packet);
  /*
   * Authenticate irn-startup-req packets
   */
  } else if(!con->query_nodename() && packet[0] == "irn-startup-req") {
    lasterror = authenticate_irn(con,packet);
  } else {
    lasterror = 0;
  }

  if(!lasterror && !con->query_nodename()) {
    IMUDDEBUG(DEBUG_CORE,sprintf("Unauthenticated packet on %s : %y",query_ip_number(con),packet));
    lasterror = ERROR_NOT_AUTHENTICATED;
  }

  if(packet[2] == "Mudfarm Lima" && packet[4] == "*wir") packet[4] = "*war";

  if(!lasterror) {
    /*
     * If this is an irn packet, check that the sending node is
     * 1. an irn node
     * 2. the originator of the packet.
     * Also check that we are the intended destination.
     */
    if(sscanf(packet[0],"irn-%*s") == 1) {
      if(!con->query_irn() || packet[2] != con->query_nodename()) {
        IMUDDEBUG(DEBUG_IRN,"irn packet from non irn peer or bad name");
        lasterror = ERROR_BAD_SOURCE;
      } else if(packet[4] != myname && packet[4] != "*") {
        IMUDDEBUG(DEBUG_IRN,"irn packet for another router");
        lasterror = ERROR_INVALID_DESTINATION;
      } else {
        lasterror = 0;
      }
    /*
     * not an irn packet, check sending mud/router against the packet
     */
    } else if(con->query_nodename() == packet[2]) {
      lasterror = 0;
    /*
     * Not any of the above, seems some mud is trying to spoof something here.
     */
    } else {
      IMUDERROR(con->query_nodename()+" is trying to spoof "+packet[2]);
      lasterror = ERROR_BAD_SOURCE;
    }
  /*
   * Allow errors even when not authenticated, they might tell us why :)
   */
  } else if(packet[0] == "error") {
    lasterror = 0;
  }

#ifdef I3_VALIDATE_PACKETS
  if(!lasterror) {
    if(PSTYPE[packet[0]]) {
      mixed * data;
      int i,m;
      data = packet[6..];
      m = sizeof(PSTYPE[packet[0]]);
      if(sizeof(data) != m) {
        lasterror = ERROR_BAD_PACKET;
        IMUDWARN("Bad packet size for "+packet[0]+" packet from "+packet[2]);
      } else {
        for(i=0;i<m;i++) {
          if(data[i] != 0 && _typeof(data[i]) != PSTYPE[packet[0]][i]) {
            lasterror = ERROR_BAD_PACKET;
            IMUDWARN("Bad type for field "+PSNAME[packet[0]][i]+" in "+packet[0]+" packet from "+packet[2]);
          }
        }
      }
    }
  }
#endif
  return !lasterror;
}

static mixed * mangle_packet(mixed * packet) {
  /*
   * Encapsulated i3 packet aimed at us (or all routers), decode.
   */
  if(packet[0] == "irn-data" && (packet[4] == "*" || packet[4] == myname)) {
    IMUDDEBUG(DEBUG_DATA,"Translating inbound irn-data packet from "+packet[2]);
    packet = packet[6];
  }
  return packet;
}

static int send_to_node(object p, mixed * data, varargs int route) {
  int r;
  int i;

  if(!route) route = 0;
  for(i=0;i<sizeof(data);i++) {
    if(data[i] == nil) error("ELEMENT "+i+" IS NILL" + sprintf("%y",data[..i-1]));
  }
  if(p == this_object() || data[4] != 0 || p->test_service_available(data[0])) {
    IMUDDEBUG(DEBUG_DATA,sprintf("send_to_node %y (%y): %y\n",p,p->query_nodename(),data));
    if(p == this_object()) {
      p->receive_router_data(data,route);
    } else if(p->query_irn()) {
      if(sscanf(data[0],"irn-%*s") != 1) {
        data = ({"irn-data",5,myname,0,"*",0,data});
      }
      p->receive_router_data(data);
    } else {
      p->receive_router_data(data);
    }
  }
}

static int broadcast_packet(mixed * data, int localsrc) {
  object * targets;
  
  targets = m_values(peers);
  if(localsrc) targets += m_values(irn);
  IMUDDEBUG(DEBUG_TRANSMIT,sprintf("Broadcasting %s packet from %s to %y",data[0],(data[3] ? data[3]+"@":"")+data[2],m_indices(peers)));
  targets = filter_array(targets,"send_to_node",this_object(),data);
  return targets && sizeof(targets);
}

static int multicast_packet(string * destination, mixed *data, int localsrc) {
  int i;
  int route;

#ifndef IRN_MULTICAST
  route = localsrc;
#endif
  IMUDDEBUG(DEBUG_TRANSMIT,sprintf("Multicasting %s from '%s' for '%s' to %y",data[0],(data[3] ? data[3]+"@":"")+data[2],data[5] ? (data[4] ? data[4]+"@":"")+data[5]:"0",destination));
  for(i=0;i<sizeof(destination);i++) {
    if(find_node(destination[i],route)) {
      send_to_node(find_node(destination[i],route),data[0..3]+({destination[i],0})+data[6..],localsrc);
    }
  }
#ifdef IRN_MULTICAST
  if(localsrc) {
    int c;
    for(c=0;c<sizeof(m_values(irn));c++) {
      if(m_values(irn)[c] != this_object()) {
        send_to_node(m_values(irn)[c],data,localsrc);
      }
    }
  }
#endif
}

static int send_packet(string destination, mixed *data, int route) {
  object target;

  target = find_node(destination,route);
  IMUDDEBUG(DEBUG_TRANSMIT,sprintf("Sending %s packet from '%s' for '%s' to '%s'",data[0],(data[3] ? data[3]+"@":"")+data[2],(data[5] ? data[5]+"@":"")+data[4],(target ? target->query_nodename():"<NONE")));

  if(target) {
    lasterror = send_to_node(target,data,route);
  } else {
    if(muddata[destination]) {
      if(query_queue_time(destination)) {
        lasterror = ERROR_TMP_DOWN;
      } else {
        lasterror = ERROR_DOWN;
      }
    } else {
      lasterror = ERROR_BAD_DESTINATION;
    }
  }
  return !lasterror;
}

static void send_error(object con, string * err, mixed * data) {
  IMUDDEBUG(DEBUG_CORE,"Sending error "+(err[0] ? ""+err[0]:"<NIL>")+":"+(err[1]? err[1]:"<NIL>")+" to "+(con->query_nodename() ? con->query_nodename():"<NOT AUTHENTICATED>"));
  con->receive_router_data( ({"error",5,myname,0,data[2],data[3],err[0],err[1],data}) );
}

int route_packet(mixed * data) {
  object target;
  int * results, localsrc;
  mixed destination;

  lasterror = 0;

  if(!(object HANDLER) previous_object() && previous_object() != this_object()) {
    error("Illegal call to route_packet()");
  }

  localsrc = previous_object()->query_localsrc() ? 1:0;
  IMUDDEBUG(DEBUG_CORE,sprintf("Packet from IP %y: %y (%y) %s",query_ip_number(previous_object()),previous_object(),previous_object()->query_nodename(),(localsrc ? "local":"irn")));
  if(!data || sizeof(data) < 6) {
    lasterror = ERROR_BAD_PACKET;
  } else if(allow_packet(previous_object(), data)) {
    if(muddata[data[2]]) muddata[data[2]]["lastseen"] = time();

    data = mangle_packet(data);
    IMUDDEBUG(DEBUG_DATA,sprintf("Processing %y",data));
    if(sscanf(data[0],"channel-%*s") == 1) {
      destination = myname;
    } else {
      destination = (data[4] ? data[4]:nil);
    }
    IMUDDEBUG(DEBUG_CORE,data[0]+" for "+(destination ? destination:"<BROADCAST>"));
    if(destination) {
      if(stringp(destination)) {
        send_packet(destination,data,localsrc);
      } else if(pointerp(destination)) {
        multicast_packet(destination,data,localsrc);
      } else {
        lasterror = ERROR_BAD_DESTINATION;
      }
    } else {
      broadcast_packet(data,localsrc);
    }
  } else {
    IMUDDEBUG(DEBUG_CORE,sprintf("Refused : %y",data));
  }

  if(lasterror && localsrc) {
    IMUDWARN(ERRORDATA[lasterror][1]+sprintf(" %y",data));
    send_error(previous_object(),query_error_data(),data+({}));
  } else if(lasterror && data[2] == previous_object()->query_nodename()) {
    IMUDWARN(ERRORDATA[lasterror][1]+sprintf("IRN: %y",data));
    send_error(previous_object(),query_error_data(),data+({}));
  } else {
    return 1;
  }
}

static int send_from_router(string ptype, mixed destmud, mixed destuser, mixed * data, varargs int route) {
  if(destmud == nil) destmud = 0;
  if(destuser == nil) destuser = 0;
  
  if(!destmud || destmud == 0) {
    broadcast_packet( ({ ptype, 5, myname, netname, 0, 0 }) + data,route);
  } else if(stringp(destmud)) {
    send_packet( destmud, ({ ptype, 5, myname, netname, destmud, destuser }) + data , route);
  } else if(pointerp(destmud)) {
    multicast_packet( destmud, ({ ptype, 5, myname, netname, 0, 0 }) + data , route);
  }
  return lasterror;
}

static int send_notice(mixed destmud, mixed destuser, mixed * data, varargs int route) {
  string ptype;

  ptype = "notice";
  if(destmud == nil) destmud = 0;
  if(destuser == nil) destuser = 0;

  IMUDDEBUG(DEBUG_SERVICES,"Sending notice to "+destuser+"@"+(arrayp(destmud) ? implode(destmud,", "):destmud));

  if(!destmud || destmud == 0) {
    broadcast_packet( ({ ptype, 5, myname, netname, 0, 0 }) + data,route);
  } else if(stringp(destmud)) {
    send_packet( destmud, ({ ptype, 5, myname, netname, destmud, destuser }) + data , route);
  } else if(pointerp(destmud)) {
    multicast_packet( destmud, ({ ptype, 5, myname, netname, 0, 0 }) + data , route);
  }
  return lasterror;
}


int query_errno() { return lasterror; }

string * query_error_data(varargs int errno) {
 if(!errno) errno = lasterror;
 if(errno >= sizeof(ERRORDATA)) errno = 0;
  return ERRORDATA[errno]; 
}

/*
 * Handle requests for the router
 */


/*
 * This is called by the low level routing process to deliver
 * data to us.
 */
static int receive_router_data(mixed data, varargs int route) {
  string ptype,func;

#ifdef IRN_DESPAM_PING
  if(data[0] != "irn-ping") {
#endif
  IMUDDEBUG(DEBUG_SERVICES,sprintf("%s packet from %s arrived at service handling",data[0],(stringp(data[3]) ? data[3]+"@":"")+data[2]));
  IMUDDEBUG(DEBUG_DATA,sprintf("%y",data));
#ifdef IRN_DESPAM_PING
  }
#endif
  ptype = data[0];
  func = router_services[ptype];
  if(func && function_exists(func,this_object())) {
#ifdef IRN_DESPAM_PING
    if(data[0] != "irn-ping")
#endif
    IMUDDEBUG(DEBUG_SERVICES,"Calling "+func);
    return call(func,data,route);
  } else {
#ifdef IRN_DESPAM_PING
    if(data[0] != "irn-ping")
#endif
    IMUDDEBUG(DEBUG_SERVICES,"Unhandled packet");
    return ERROR_BAD_PACKET;
  }
}

static int test_connected(string mud) {
  return muddata[mud] && 
         muddata[mud]["connect_time"] && 
         !muddata[mud]["disconnect_time"] &&
         (muddata[mud]["router"] && (muddata[mud]["router"] == myname || irn[muddata[mud]["router"]])) ;
}

/*
 * Reply to a startup request
 */
static int rcv_startup_req(mixed * data, varargs int route) {
  string mud;
  int version;

  mud = (data[2] ? data[2]:nil);
  if(!mud || sizeof(data) < 19) {
    return ERROR_BAD_PACKET;
  }

  IMUDDEBUG(DEBUG_SERVICES,"sending startup reply to "+mud);

  sscanf(data[0],"startup-req-%d",version);
  find_node(mud)->set_protocol_version(version);
  send_from_router("startup-reply",mud,0,({filtered_routerlist(),muddata[mud]["password"]}));

  if(!lasterror) {
    if(mappingp(data[18])) {
      find_node(mud)->set_services(data[18]);
    }
    if(mappingp(data[19])) {
      if(data[19]["imud_version"]) {
        version=data[19]["imud_version"];
        find_node(mud)->set_protocol_version(version);
      }
    }

/*
    if(query_queue_time(mud)) {
      dequeue_mudlist_changed(mud);
      irn_broadcast_mudlist_change(mud);
    } else {
      mudlist_changed(mud,1,1);
    }
*/

    call_out("send_mudlist",0,mud,data[7]);
    call_out("send_chanlist",0,mud);
    send_notice(({"Way of the Force","The Zone","WOTF Devel"}),0,({mud+" connected from "+muddata[mud]["lastip"]+" using the I3 version "+version+" protocol"}),1);
    return 0;
  }
  return lasterror;
}

static int rcv_shutdown(mixed * data, varargs int route) {
  string mud;

  if(sizeof(data) != 7) {
    return ERROR_BAD_PACKET;
  }

  if(peers[data[2]]) {
    if(muddata[data[2]]) {
      muddata[data[2]]["delay"] = data[6];
    }
    peers[data[2]]->force_disconnect();
  }
}

static int rcv_mudlist_req(mixed * data, varargs int route) {
  IMUDDEBUG(DEBUG_SERVICES,"rcv_mudlist_req");
  return send_mudlist(data[2], muddata[data[2]]["public"][0]);
}

static int test_allow_on_channel(string mud, string channel) {
  switch(chandata[channel][1]) {
    case 0 : return member_array(mud,chandata[channel][2]) == -1;
    case 1 :
    case 2 : return member_array(mud,chandata[channel][2]) != -1;
  }
}

static int send_irn_chanlist_delta(string channel, varargs int nochan, int nolisten) {
  mapping data;
  int i;
  object * router;

  data = ([]);
  if(!nochan) {
    data["channels"] = ([ channel : (chandata[channel] ?  ({ chandata[channel][1], chandata[channel][0], chandata[channel][2] }):0) ]);
  }
  if(!nolisten) {
    data["listening"] = ([ channel : (chandata[channel] ? chandata[channel][3]:({})) ]);
  }

  router = m_values(irn);
  for(i=0;i<sizeof(router);i++) {
    send_to_node(
      router[i], 
      ({
        "irn-chanlist-delta",
        5,
        myname,
        0,
        router[i]->query_nodename(),
        0,
        chanlistid,
        data
      })
    );
  }
}

static int rcv_irn_ping(mixed * data, varargs int route) {
#ifndef IRN_DESPAM_PING
  IMUDDEBUG(DEBUG_IRN,"Received irn-ping from "+data[2]); */
#endif
}

static int rcv_irn_chanlist_delta(mixed * data, varargs int route) {
  int i,j,k;
  int irnc,pubc;

  string * tmp;
  mixed channels;
  mixed listeners;

  IMUDDEBUG(DEBUG_SERVICES,"rcv_irn_chanlist_delta");

  if(!mappingp(data[7])) {
    return ERROR_BAD_PACKET;
  }

  channels = data[7]["channels"];
  listeners = data[7]["listening"];
  chanlistid = data[6];

  if(channels) {
    IMUDDEBUG(DEBUG_SERVICES,"We have channels: "+implode(m_indices(channels),", "));
    if(!mappingp(channels)) {
      return ERROR_BAD_PACKET;
    }

    pubc = 1;
    tmp = m_indices(channels);
    for(i=0;i<sizeof(tmp);i++) {
      if(channels[tmp[i]] == 0) {
        chandata[tmp[i]] = nil;
      } else if(!chandata[tmp[i]]) {
        chandata[tmp[i]] = ({ channels[tmp[i]][1],
                              channels[tmp[i]][0],
                              channels[tmp[i]][2],
                              ({ })
                           });
      } else {
        chandata[tmp[i]] = ({ channels[tmp[i]][1],
                              channels[tmp[i]][0],
                              channels[tmp[i]][2],
                              chandata[tmp[i]][3]
                            });
      }
    }
  }

  if(listeners) {
    IMUDDEBUG(DEBUG_SERVICES,"We have listeners");
    if(!mappingp(listeners)) {
      return ERROR_BAD_PACKET;
    }
    tmp = m_indices(listeners);
    for(i=0;i<sizeof(tmp);i++) {
      if(!chandata[tmp[i]]) {
        /* can't add listeners to a channel we don't have */
        /* chandata[tmp[i]] = ({ "The Zone",0,({}),({}) }); */

        IMUDWARN("listeners received for "+tmp[i]+" but I don't have that channel!");
      } else {
        IMUDDEBUG(DEBUG_SERVICES,"Changing listeners for "+tmp[i]+" to "+sprintf("%y",listeners[tmp[i]]));
        chandata[tmp[i]][3] = listeners[tmp[i]];
      }
    }
  }
  save_me();
  if(pubc) {
    filter_array(m_indices(channels),"broadcast_chanlist",this_object());
  }
}



/*
 * Handle channel requests
 */
static int rcv_channel(mixed * data, varargs int route) {
  string cname, *destinations;
  int i;

  if(sizeof(data) < 7) {
    IMUDDEBUG(DEBUG_CHANNELS,"Bad packet (bad size)");
    return ERROR_BAD_PACKET;
  }

  cname = data[6] ? data[6]:nil;
  if(!cname) {
    IMUDDEBUG(DEBUG_CHANNELS,"Bad packet (no channel name)");
    return ERROR_BAD_PACKET;
  }

  if(!chandata[cname]) {
    IMUDDEBUG(DEBUG_CHANNELS,"Unknown channel");
    send_from_router("error",data[2],data[3],({"unk-channel","The channel "+cname+" is not known here",data}));
  } else if(!test_allow_on_channel(data[2],cname)) {
    IMUDDEBUG(DEBUG_CHANNELS,"Not allowed");
    send_from_router("error",data[2],data[3],({"not-allowed","Your mud is not allowed to use the "+cname+" channel",data}));
  } else if(chandata[cname]) {
    IMUDDEBUG(DEBUG_CHANNELS,"Trying to deliver");
    destinations = chandata[cname][3];
    if(destinations && sizeof(destinations)) {
       IMUDDEBUG(DEBUG_CHANNELS,"Delivering "+data[0]+" to "+sprintf("%y",destinations));
       multicast_packet(destinations,data,0);
    } else {
      IMUDDEBUG(DEBUG_CHANNELS,"No listeners?");
    }

    if(route) {
      for(i=0;i<m_sizeof(irn);i++) {
        send_to_node(m_values(irn)[i], data);
      }
    }
  }
}

static int rcv_chan_adminlist(mixed * data, varargs int route) {
  mixed src, user, cname;

  if(sizeof(data) != 7) {
    return ERROR_BAD_PACKET;
  }

  src = data[2];
  user = data[3];
  cname = data[6];

  if(!chandata[cname]) {
    send_from_router("error",src,user,({"unk-channel","The channel "+cname+" is not known here.",data }) );
    return 0;
  }

  if(chandata[cname][0] != src) {
    send_from_router("error",src,user,({"not-allowed","Only the mud owning "+cname+" is allowed to do that", data}) );
    return 0;
  }

  return send_from_router("chan-adminlist-reply",src,user,({ cname, chandata[cname][2] }) );
  send_irn_chanlist_delta(cname);
}

static int rcv_channel_listen(mixed * data, varargs int route) {
  string src,cname;
  int flag;

  if(sizeof(data) != 8) {
    return ERROR_BAD_PACKET;
  }

  src = (data[2] ? data[2]:nil);
  cname = (data[6] ? data[6]:nil);
  if(!cname || !src) {
    return ERROR_BAD_PACKET;
  }
  flag = data[7];
  if(!chandata) chandata = ([]);

  if(chandata[cname]) {
    chandata[cname][3] -= ({src});
    if(flag) {
      if(test_allow_on_channel(src,cname)) {
        if(member_array(src,chandata[cname][3]) == -1) {
          chandata[cname][3] += ({src});
        }
        save_me();
        send_irn_chanlist_delta(cname,1,0);
      } else {
        send_from_router("error",src,data[3],({"not-allowed","Your mud is not allowed to use that channel",data}));
      }
    }
  }
}

static int rcv_channel_add(mixed * data, varargs int route) {
  string src,cname;
  int mode;

  if(!chandata) {
    chandata = ([]);
  }

  if(sizeof(data) != 8) {
    return ERROR_BAD_PACKET;
  }

  src = data[2];
  cname = (data[6] ? data[6]:nil);
  mode = data[7];

  if(!cname) {
    return ERROR_BAD_PACKET;
  }

  if(chandata[cname]) {
    send_from_router("error",src,data[3],({"not-allowed","The channel "+cname+" already exists.",data}));
  } else {
    IMUDLOG("Channel "+cname+" added by "+((data[3] && data[3]) != 0 ? data[3]:"<NONE>")+"@"+src);
    /* format: "channelname":({"owning mud",mode,({banlist}),({listeners})}) */
    chandata[cname] = ({src,mode,({}),({})});
    chanlistid++;
    save_me();
    call_out("broadcast_chanlist",0,cname);
    send_irn_chanlist_delta(cname);
  }
}

static int rcv_channel_remove(mixed * data, varargs int route) {
  string src,cname;
  int mode;

  if(!chandata) {
    chandata = ([]);
  }

  if(sizeof(data) != 7) {
    return ERROR_BAD_PACKET;
  }

  src = data[2];
  cname = (data[6] ? data[6]:nil);

  if(!cname) {
    return ERROR_BAD_PACKET;
  }

  if(!chandata[cname]) {
    send_from_router("error",src,data[3],({"unk-channel","The channel "+cname+" does not exist here.",data}));
  } else if(chandata[cname][0] != src) {
    send_from_router("error",src,data[3],({"not-allowed","Your mud does not own the channel "+cname+".",data}));
  } else {
    IMUDLOG("Channel "+cname+" removed by "+((data[3] && data[3]) != 0 ? data[3]:"<NONE>")+"@"+src);
    chandata[cname] = 0;
    chanlistid++;
    save_me();
    call_out("broadcast_chanlist",0,cname);
    send_irn_chanlist_delta(cname);
  }
}

static int rcv_channel_admin(mixed * data, varargs int route) {
  string src, cname;

  if(sizeof(data) != 9) {
    return ERROR_BAD_PACKET;
  }

  src = data[2];
  cname = (data[6] ? data[6]:nil);

  if(!cname) {
    return ERROR_BAD_PACKET;
  }

  if(!chandata[cname]) {
    send_from_router("error",src,data[3],({"unk-channel","The channel "+cname+" is not known here.",data}));
  } else if(chandata[cname][0] != src) {
    send_from_router("error",src,data[3],({"not-allowed","Your mud does not own the channel "+cname+".",data}));
  } else {
    IMUDLOG("List for "+cname+" changed by "+((data[3] && data[3]) != 0 ? data[3]:"<NONE>")+"@"+src);
    if(data[7]) chandata[cname][2] += data[7];
    if(data[8]) chandata[cname][2] -= data[8];
    chanlistid++;
    save_me();
    call_out("broadcast_chanlist",0,cname);
    send_irn_chanlist_delta(cname);
  }
}

static int rcv_error(mixed * data, varargs int route) {
}

static int send_chanlist(string where, varargs string cname) {
  string * channels;
  int i;
  mapping result;

  if(!chandata) chandata = ([]);
  result = ([]);
  IMUDDEBUG(DEBUG_SERVICES,"Sending channel list for "+(cname ? cname:"all channels")+" to "+where);
  if(!cname) {
    channels = m_indices(chandata);
  } else {
    channels = ({ cname });
  }
  for(i=0;i<sizeof(channels);i++) {
    if(chandata[channels[i]] != nil) {
      result[channels[i]] = chandata[channels[i]][..1];
    } else {
      result[channels[i]] = 0;
    }
  }
  if(result && m_sizeof(result)) {
    send_from_router( "chanlist-reply", where, 0, ({ chanlistid, result }) );
  }
}

static int check_channel(mixed stuff) {
  return stuff != 0;
}

static int broadcast_chanlist(varargs string cname) {
  string * destinations;
  destinations = m_indices(peers);
  if(destinations && sizeof(destinations)) filter_array(destinations,"send_chanlist",this_object(),cname);
  chandata = filter_mapping(chandata,"check_channel",this_object());
}

static void notify_mud_removed(string src) {
  string * channels;
  int i;

  channels = m_indices(chandata);
  for(i=0;i<sizeof(channels);i++) {
    chandata[channels[i]][3] -= ({src});
  }
}


static int mark_for_router_down(string mud, string router, varargs int flag) {
  if(muddata[mud] && muddata[mud]["router"] == router && muddata[mud]["delay"] == -1) {
    muddata[mud]["delay"] = 0;
    return 1;
  }
}

void notify_disconnect() {
  string mudname;

  if(!(object HANDLER) previous_object()) {
    error("Illegal call to notify_disconnect()");
  }

  mudname = previous_object()->query_nodename();
  IMUDLOG("Mud "+mudname+" disconnected.");
  if(peers[mudname] && muddata[mudname]["router"] == myname) {
    peers[mudname]=nil;
    muddata[mudname]["connect_time"] = 0;
    muddata[mudname]["disconnect_time"] = time();

    send_notice(({"Way of the Force","The Zone","WOTF Devel"}),0,({mudname+" disconnected."}),1);

    if(muddata[mudname]["delay"] == -1) {
      IMUDDEBUG(DEBUG_CORE,"notify_disconnect::delay == -1, setting to 0");
      muddata[mudname]["delay"] = 0;
      queue_mudlist_changed(mudname);
      irn_broadcast_mudlist_change(mudname);
    } else if(muddata[mudname]["delay"] > (7*24*3600)) {
      IMUDDEBUG(DEBUG_CORE,"notify_disconnect::delay > 7 weeks, removing");
      muddata[mudname] = nil;
      mudlist_changed(mudname,1,1);
    } else if(!muddata[mudname] || muddata[mudname]["delay"] >= 300) {
      IMUDDEBUG(DEBUG_CORE,"notify_disconnect::delay > 5 minutes, inmediately notify the world.");
      mudlist_changed(mudname,1,1);
    } else {
      IMUDDEBUG(DEBUG_CORE,"notify_disconnect::delay > 0 and < 5 mins");
      queue_mudlist_changed(mudname);
      irn_broadcast_mudlist_change(mudname);
    }
  }
}

static int rcv_irn_mudlist_delta(mixed * data, varargs int route) {
  mapping ml;
  mapping md;
  string * muds;
  int i;

  IMUDDEBUG(DEBUG_IRN,"IRN mudlist delta received from " +data[2]);
  ml = data[7];
  muds = m_indices(ml);
  for(i=0;i<sizeof(muds);i++) {
    if(ml[muds[i]] == 0) {
      if(muddata[muds[i]]) {
        IMUDDEBUG((DEBUG_IRN | DEBUG_SERVICES),"Removing mud "+muds[i]);
        muddata[muds[i]] = nil;
        notify_mud_removed(muds[i]);
        mudlist_changed(muds[i],0,0);
      } else {
        IMUDDEBUG((DEBUG_IRN | DEBUG_SERVICES),data[2]+" tried to remove non existing mud "+muds[i]);
      }
    } else {
      md = ml[muds[i]];
      if(md["connect_time"] == 0 && md["restart_delay"] == -1) {
        IMUDDEBUG(DEBUG_SERVICES,"Fixing restart_delay of -1 for offline mud "+muds[i]);
        md["restart_delay"] = 0;
      }

      if(md["router"] == data[2]) {
        if(muddata[muds[i]]) {
          if(intp(md["restart_delay"]) && md["restart_delay"] >= 0 && md["router"] != muddata[muds[i]]["router"]) {
            IMUDDEBUG((DEBUG_IRN | DEBUG_SERVICES),data[2]+" reports "+muds[i]+" offline but it is currently connected to "+muddata[muds[i]]["router"]);
          } else {
            IMUDDEBUG((DEBUG_IRN | DEBUG_SERVICES), "Changing "+" mud "+muds[i]);
            store_irn_muddata(md);
            irnmudlistid = data[6];
          }
        } else {
          IMUDDEBUG((DEBUG_IRN | DEBUG_SERVICES), "Adding "+" mud "+muds[i]);
          store_irn_muddata(md);
          irnmudlistid = data[6]; 
        }
      } else {
        IMUDDEBUG((DEBUG_IRN | DEBUG_SERVICES),data[2]+" is not currently responsible for "+muds[i]);
      }
    }
  }
}

static int rcv_irn_startup_req(mixed * data, varargs int route) {
  string peername;

  IMUDDEBUG(DEBUG_IRN,"IRN startup request received!");
  peername = data[2];
  send_irn_mudlist(peername);
  send_notice(({"Way of the Force","The Zone","WOTF Devel"}),0,({"Router "+peername+" connected."}),1);
}

static int rcv_irn_data(mixed * data, varargs int route) {
#ifdef IRNV2
  object r;

  if(data[1]) {
    r = find_irn_route(data[2]);
    if(r) {
      data[1]--;
      send_to_node(r,data);
    }
  }
#else
  IMUDWARN(sprintf("Received irn-data for another router: %y\n",data));
#endif
}

object _Q_connection() { return this_object(); }

static int filter_muddata(mapping data) {
  return !(data["disconnect_time"]) || (data["disconnect_time"] == 0 || (time() - data["disconnect_time"] <= (14 * 24 * 3600))); 
}

static void expire_pending() { }
    
static void periodic() {
  call_out("periodic",60);
  if(time() - lastclean >= 3600) {
    mudlist::do_cleanup();
    lastclean = time();
  }
  expire_pending();
  run_queue();
  irn_reconnect();
}

void upgraded() { 
}

void remove_mud(string str) {
  if(peers[str]) peers[str]->force_disconnect();
  ::remove_mud(str);
}

void init_routerlist() {
  routerlist = DEFAULT_ROUTERS;
  save_me();
}

void enable_irn_node(string node) {
  if(routerlist[node]) {
    routerlist[node][3] = 1;
    save_me();
  }
}

void disable_irn_node(string node) {
  if(routerlist[node]) {
    routerlist[node][3] = 0;
    if(irn[node]) irn[node]->force_disconnect();
    save_me();
  }
}

void set_mypass() { mypass = MYPASS; }

string query_local_muds() {
  return implode(m_indices(peers),", ")+".\n";
}

void add_expired(string str) {
  expired[str] = mudlistid;
}
mapping query_muddata() { return muddata; }
string * query_expired() { return m_indices(expired); }
string * query_listeners(string str) {
  if(chandata[str]) {
    return chandata[str][3];
  }
}

void resend_startup_reply(string mud) {
  if(peers[mud]) {
    send_from_router("startup-reply",mud,0,({filtered_routerlist(),muddata[mud]["password"]}));
  }
}

string * query_irn_connections() { return m_indices(irn); }

string query_netname() {
  if(netname) return netname;
  else return "<NONE>";
}

string * query_local_mud_names() {
  return m_indices(peers);
}

string * query_irn_names() {
  return m_indices(irn);
}


