#ifndef ROUTER_DOT_H_INCLUDED
#define ROUTER_DOT_H_INCLUDED

#include <routercfg.h>

/*
 * debug bits
 */
#define DEBUG_CORE     1
#define DEBUG_TRANSMIT 2
#define DEBUG_DATA     4
#define DEBUG_AUTH     8
#define DEBUG_CHANNELS 16
#define DEBUG_SERVICES 32
#define DEBUG_IRN      64

/*
 * Debugging/logging macros
 */
#define IMUDDEBUG(L,X) if(DBL & L) log_file("router",X+"\n",LOGLEVEL_DEBUG)
#define IMUDWARN(X)  log_file("router",X+"\n",LOGLEVEL_WARN)
#define IMUDERROR(X) log_file("router",X+"\n",LOGLEVEL_ERROR)
#define IMUDLOG(X)   log_file("router",X+"\n",LOGLEVEL_NOTICE)

/*
 * some filenames
 */
#define COMMS "/system/daemon/comms"
#define HANDLER "/system/obj/net/I3_inb_server"
#define DRIVER "/kernel/sys/driver"
#define SAVEFILE "/system/save/I3-router"

/*
 * the list of services is defined here..
 */
#define router_services ([\
  "startup-req-1":"rcv_startup_req",\
  "startup-req-2":"rcv_startup_req",\
  "startup-req-3":"rcv_startup_req",\
  "startup-req-4":"rcv_startup_req",\
  "mudlist-req":"rcv_mudlist_req",\
  "shutdown":"rcv_shutdown",\
  "channel-add":"rcv_channel_add",\
  "channel-remove":"rcv_channel_remove",\
  "channel-admin":"rcv_channel_admin",\
  "chan-adminlist":"rcv_chan_adminlist",\
  "channel-m":"rcv_channel",\
  "channel-e":"rcv_channel",\
  "channel-t":"rcv_channel",\
  "channel-listen":"rcv_channel_listen",\
  "chan-filter-reply":"rcv_chan_filter_reply",\
  "error":"rcv_error",\
  "irn-startup-req":"rcv_irn_startup_req",\
  "irn-startup-reply":"rcv_irn_startup_reply",\
  "irn-mudlist-delta":"rcv_irn_mudlist_delta",\
  "irn-mudlist-altered":"rcv_irn_mudlist_delta",\
  "irn-chanlist-delta":"rcv_irn_chanlist_delta",\
  "irn-chanlist-altered":"rcv_irn_chanlist_delta",\
  "irn-ping-req":"rcv_irn_ping",\
  "irn-ping":"rcv_irn_ping",\
  "irn-shutdown":"rcv_irn_shutdown",\
])

/*
 * prototypes
 */

static void    mudlist_changed(string what, varargs int flag, int route);
       string  *query_error_data(varargs int errno);
       int     broadcast_mudlist_change(string what, varargs string * skip);
static int     broadcast_packet(mixed * data, int localsrc);
static int     multicast_packet(string * destination, mixed *data, int localsrc);
static int     send_packet(string destination, mixed *data, int localsrc);
static int     send_to_node(object p, mixed * data, varargs int route);
static void    irn_broadcast_mudlist_change(string what);
static void    notify_mud_removed(string src);
static int     send_from_router(string ptype, mixed destmud, mixed destuser, mixed * data, varargs int route);
       object  find_irn(string str);
static mapping i3_nodes();
static string  *i3_node_names();
static object  *i3_node_objects();
static mapping irn_nodes();
static string  *irn_node_names();
static object  *irn_node_objects();


#endif
