/*
 * IRN related functionality
 *
 */

#include <intermud3.h>
#include <router.h>
#include <loglevels.h>

static mapping irn;
private int DBL;

static mapping query_router_list();
static string query_mypass();
static string query_myname();
static int send_notice(mixed destmud, mixed destuser, mixed * data, varargs int route);
mapping query_muddata();

static void set_irn_dbl(int i) {
  DBL = i;
}

static mapping irn_nodes() {
  return irn;
}

static string *irn_node_names() {
  return m_indices(irn);
}

static object *irn_node_objects() {
  return m_values(irn);
}

void start_irn_con(string irnname) {
  string ip;
  int port;
  string srvpass;
  object con;
  mapping rl;

  rl = query_router_list();

  if(rl[irnname] && !irn[irnname] && rl[irnname][3] == 1) {
    IMUDDEBUG(DEBUG_IRN,"Trying to connect to "+(irnname ? irnname:"<NONE>"));
    ip = rl[irnname][0];
    port = rl[irnname][1];
    srvpass = rl[irnname][2];
    con = clone_object(HANDLER);
    con->irn_outgoing(irnname,query_mypass(),srvpass,ip,port);
  } else {
    IMUDDEBUG(DEBUG_IRN,"Skipping "+(irnname ? irnname:"<NONE>")+", it is disabled.");
  }
}

void stop_irn_con(string str) {
  if(irn[str]) {
    irn[str]->force_disconnect();
  }
}

static void register_irn(object o, string n) {
  if(!irn) irn = ([]);
  o->set_nodename(n);
  o->set_irn(1);
  o->set_localsrc(0);
  irn[n] = o;
  IMUDLOG("irn node "+n+" connected from "+query_ip_number(o));
}

static int authenticate_irn(object con, mixed *data) {
  string src,dst,ip;
  mapping rl;

  IMUDDEBUG((DEBUG_IRN | DEBUG_AUTH),"IRN authentication request from "+query_ip_number(con));
  if(sizeof(data) < 7 || !mappingp(data[6])) return ERROR_BAD_PACKET;
  src = (data[2] ? data[2]:nil);
  dst = (data[4] ? data[4]:nil);
  IMUDDEBUG((DEBUG_IRN | DEBUG_AUTH),"Claims to be "+(src ? src:"<NO NAME>"));
  if(!src) {
    return ERROR_BAD_SOURCE;
  }

  if(dst != query_myname()) {
    IMUDDEBUG((DEBUG_IRN | DEBUG_AUTH),"bad destination for auth packet: "+dst);
    return ERROR_INVALID_DESTINATION;
  }

  rl = query_router_list();

  if(!rl[src]) {
    IMUDDEBUG((DEBUG_IRN | DEBUG_AUTH),"Unknown router: "+src);
    return ERROR_PERMISSION_DENIED;
  }

  ip = query_ip_number(con);
  if(rl[src][0] != ip) {
    IMUDDEBUG((DEBUG_IRN | DEBUG_AUTH),"Bad ip "+ip+" for "+src+", it is registered at "+rl[src][0]);
    return ERROR_PERMISSION_DENIED;
  }
  if(data[6]["server_password"] != query_mypass() ||
     data[6]["client_password"] != rl[src][2]
  ) {
    IMUDDEBUG((DEBUG_IRN | DEBUG_AUTH),"Bad password for "+src);
    return ERROR_PERMISSION_DENIED;
  }

  /*
   * Allow replacing dead connections
   * Also, discard spurious registation attempts.
   */
  if(irn[src] && irn[src] != con) {
    IMUDDEBUG(DEBUG_IRN,"Forcing disconnect of other session: "+object_name(irn[src]));
    irn[src]->force_disconnect();
  }
  if(!irn[src]) register_irn(con,src);
}

object find_irn(string str) {
  return irn[str];
}

int register_outbound_irn() {
  object con;
  string src;

  con = (object HANDLER) previous_object();
  if(!con) return 0;

  src = con->query_nodename();
  /*
   * If there already exists a connection to the claimed name,
   * refuse to register
   */
  if(find_irn(src)) {
    return 0;
  } else {
    register_irn(con,src);
    call_out("send_irn_mudlist",0,src);
    return 1;
  }
}

static void irn_reconnect() {
  int i;
  string *nodes;
  string myname;

  myname = query_myname();

  nodes = m_indices(query_router_list());
  for(i=0;i<sizeof(nodes);i++) {
    if(nodes[i] != myname && !irn[nodes[i]]) {
      start_irn_con(nodes[i]);
    } else if((nodes[i] != myname) && irn[nodes[i]] && (time() - irn[nodes[i]]->query_lastread() >= 120)) {
      stop_irn_con(nodes[i]);
    } else if((nodes[i] != myname) && irn[nodes[i]] && (time() - irn[nodes[i]]->query_lastsent() >= 60)) {
      send_to_node(irn[nodes[i]],({"irn-ping",5,myname,0,nodes[i],0,time(),time()}));
    }
  }
}

void notify_irn_disconnect() {
  string routername;
  string *irnmuds;

  if(!(object HANDLER) previous_object()) {
    error("Illegal call to notify_irn_disconnect()");
  }

  routername = previous_object()->query_nodename();
  if(routername) {
    IMUDWARN("Router "+routername+" disconnected.");
    irn[routername]=nil;
    irnmuds = m_indices(query_muddata());
    irnmuds = filter_array(irnmuds,"mark_for_router_down",this_object(),routername);
    irnmuds = filter_array(irnmuds,"queue_mudlist_changed",this_object());
    send_notice(({"Way of the Force","The Zone","WOTF Devel"}),0,({"Router "+routername+" disconnected."}),1);
  }
}

