/*
  I3 inband server object
  This should be used together with an I3 router object

  This provides a connection for a single mud.
*/

#ifdef status
#undef status
#endif

#include <config.h>
#include <status.h>
#include <limits.h>
#include <routercfg.h>
#include <intermud3.h>
#include <loglevels.h>

#define DBT(X) if(find_player("aidil")) find_player("aidil")->tell(X+"\n")
#define ROUTER "/system/daemon/net/I3_router"


inherit "/system/lib/net/services";
inherit "/kernel/lib/net/netobj";
inherit "/lib/timefuns";

static int connected, closing, down, lastsent, lastread, pktsent, pktread;
static int trace_protocol;
string netname;
static mixed * packets;
static mapping services;
static int protocol_version;
static int localsrc;
static int irn;

void set_protocol_version(int v) {
  protocol_version = v;
}

int query_protocol_version() {
  return protocol_version;
}

void set_services(mapping s) {
  if(mappingp(s)) {
    services = s + ([]);
  }
}

mapping query_services() {
  return services ? services + ([]) : ([]);
}

#ifdef I3_UCACHE_FILTERING
int test_service_available(string service) {
  if(irn) return 1;
  switch(service) {
    case "ucache-update" : 
                             return services && services["ucache"];
                             break;
    case "router-shutdown" : 
                             return (protocol_version == 4);
                             break;
    default                : 
                             return 1;
                             break;
  }
}
#else
int test_service_available(string service) {
  IMUDLOG("huh?");
  return 1 || (services && services[service] == 1);
}
#endif

void set_nodename(string str) {
  connected = 1;
  netname = str;
}

string query_nodename() {
  return netname;
}

int query_trace_protocol() { return trace_protocol; }
void set_trace_protocol(int i) { trace_protocol = (i != 0); }

void IMUDLOG( string str ) {
  log_file( "router","("+netname+") : " + str ,LOGLEVEL_NOTICE);
}

void IMUDERROR(string str) {
  log_file( "router", "("+netname+") : "+str, LOGLEVEL_ERROR);
}

void TRACEIN(mixed data) {
  if(trace_protocol) {
    write_file("/log/admin/intermud3_trace",
      "<< "+ctime(time())+"\n"+
      sprintf("%y\n",data)
    );
  }
}

void TRACEOUT(mixed data) {
  if(trace_protocol) {
    write_file("/log/admin/intermud3_trace",
      ">> "+ctime(time())+"\n"+
      sprintf("%y\n",data)
    );
  }
}

static void send_loop() {
  if(packets && sizeof(packets)) {
    mixed packet;
    packet = packets[0];
    if(sizeof(packets) > 1) {
      packets = packets[1..];
      call_out("send_loop",0);
    } else {
      packets = ({});
    }
    pktsent++;
    lastsent = time();
    send_message(packet);
  }
}

mixed * mangle_incoming_packet(mixed * data) {
  switch(data[0]) {
    case "startup-req" : log_file("router","translating startup-req\n",LOGLEVEL_DEBUG);
    case "startup-req-1" : log_file("router","version 1\n",LOGLEVEL_DEBUG);
      data = data[0..16]+({""})+data[17..]+({0});
      data[0] = "startup-req-1";
  }
  return data;
}
    
/* 
 * translate outgoing packets for the target mud if needed.
 */
mixed * mangle_outgoing_packet(mixed * data) {
  mixed * result;
  if(protocol_version == 0 || protocol_version == 4) {
    result = data;
  } else {
    switch(data[0]) {
      case "mudlist-reply" :
        result = ({"mudlist"})+data[1..];
        break;
      case "mudlist-delta" :
        result = ({"mudlist",5})+data[2..6]+({([data[7]:data[8]])});
        break;
      case "chanlist-delta" :
        result = ({"chanlist-reply",5})+data[2..6]+({([data[7]:data[8]])});
        break;
      default :
        result = data;
        break;
    }
  }
  if(protocol_version == 1 && result[0] == "mudlist") {
    int i,j;
    mixed mud;
    string *muds;
    muds = m_indices(result[7]);
    for(i=0;i<sizeof(muds);i++) {
      mud = result[7][muds[i]];
      if(arrayp(mud)) {
        mud = mud[0..9] + mud[11..11];
        result[7][muds[i]] = mud;
      }
    }
  }  
  return result;
}

public int receive_router_data(mixed * data) {
  if(!packets) packets = ({});
  if(data) {
    data = mangle_outgoing_packet(data);
    packets += ({ data });
    if(find_call_out("send_loop") == -1) {
      call_out("send_loop",0);
    }
  }
}


static void send_router_data(mixed data) {
  pktread++;
  lastread = time();
  data = mangle_incoming_packet(data);
  ROUTER->route_packet(data);
}

void receive_message(mixed data) {
  if(data && pointerp(data)) send_router_data(data);
  else if(data) IMUDERROR(sprintf("%y",data));
}

void close() {
  if(closing) return;
  closing = 1;
  if(previous_object() != _Q_connection() && previous_object() != this_object()) return;
  catch(remove_call_out("timeout_connection"));
  catch(destruct(this_object()));
}

void timeout_connection() {
  if(connected) return;
  if(catch(close_user())) {
    close();
  }
}

static int outgoing;
static string mypass;
static string theirpass;

void irn_outgoing(string iname, string myp, string theirp, string ip, int port) {
  set_nodename(iname);
  mypass = myp;
  theirpass = theirp;
  outgoing = 1;
  connect(ip,port,"mudmode");
}

void open()
{
  if(outgoing) {
    receive_router_data(
      ({"irn-startup-req",
        5,
        ROUTER->query_nodename(),
        0,
        query_nodename(),
        0,
        (["client_password":mypass,
          "server_password":theirpass
        ])
      })
    );
    if(ROUTER->register_outbound_irn()) {
      connected = 1;
    } else {
      call_out("timeout_connection",0);
    }
  }
}

void destroying() {
  catch(close_user());
  if(irn) catch(ROUTER->notify_irn_disconnect());
  else catch(ROUTER->notify_disconnect());
}

void create() {
  if(!original()) {
    call_out("timeout_connection",120);
    packets = ({});
  }
}

void force_disconnect() {
  if(previous_object() == get_object(ROUTER)) {
    if(!closing) {
      closing = 1;
      catch(destruct_object(this_object()));
    }
  }
}

void set_localsrc(int flag) {
  localsrc = flag ? 1:0;
}

void set_irn(int flag) {
  irn = flag ? 1:0;
}

int query_localsrc() { return localsrc; }
int query_irn() { return irn; }

void upgraded() {
  if(!original() && !connected) {
    call_out("timeout_connection",10);
  }
}

int query_pktsent() { return pktsent; }
int query_pktread() { return pktread; }
int query_lastsent() { return lastsent; }
int query_lastread() { return lastread; }

